#include <stdarg.h>	/* va_* */
#include <stdio.h>	/* *printf */
#include <stdlib.h>	/* getenv */
#include <string.h>	/* strcasecmp */

#define BUF_SIZE	256

enum {
	LEVEL_ERROR,
	LEVEL_WARN,
	LEVEL_INFO,
	LEVEL_DEBUG,
};
#define LEVEL_DEFAULT	LEVEL_INFO



static inline int
get_level(void)
{
	const char *level = getenv("CONDNET_LOG");
	if (level == NULL) {
		return LEVEL_DEFAULT;
	} else if (strcasecmp(level, "DEBUG") == 0) {
		return LEVEL_DEBUG;
	} else if (strcasecmp(level, "INFO") == 0) {
		return LEVEL_INFO;
	} else if (strcasecmp(level, "WARN") == 0) {
		return LEVEL_WARN;
	} else if (strcasecmp(level, "ERROR") == 0) {
		return LEVEL_ERROR;
	} else {
		return LEVEL_DEFAULT;
	}
}


static inline int
show_level(int level)
{
	return level <= get_level();
}



static inline void
log_level(const char *level, const char *fmt, va_list va)
{
	char buf[BUF_SIZE + 1];
	vsnprintf(buf, BUF_SIZE, fmt, va);
	buf[BUF_SIZE] = '\0';
	if (level) {
		fprintf(stderr, "[%4s] %s\n", level, buf);
	} else {
		fprintf(stderr, "%s\n", buf);
	}
}



void
log_debug(const char *fmt, ...)
{
	va_list va;
	if (!show_level(LEVEL_DEBUG)) {
		return;
	}
	va_start(va, fmt);
	log_level("DBUG", fmt, va);
	va_end(va);
}



void
log_info(const char *fmt, ...)
{
	va_list va;
	if (!show_level(LEVEL_INFO)) {
		return;
	}
	va_start(va, fmt);
	log_level("INFO", fmt, va);
	va_end(va);
}



void
log_warn(const char *fmt, ...)
{
	va_list va;
	if (!show_level(LEVEL_WARN)) {
		return;
	}
	va_start(va, fmt);
	log_level("WARN", fmt, va);
	va_end(va);
}



void
log_error(const char *fmt, ...)
{
	va_list va;
	if (!show_level(LEVEL_ERROR)) {
		return;
	}
	va_start(va, fmt);
	log_level("ERR", fmt, va);
	va_end(va);
}
