#include "rules.h"
#include "log.h"

#include <string.h>	/* strncpy, strtok_r, memset */
#include <stdlib.h>	/* malloc */
#include <assert.h>	/* assert */
#include <errno.h>	/* ETIMEDOUT */

#define RULE_LEN	256


enum {
	TYPE_INVALID,
	TYPE_CONNECTION,
	TYPE_CONNECTION_INBOUND,
	TYPE_CONNECTION_OUTBOUND,
	TYPE_RESOLVE,
	TYPE_POLICY,
};

struct policy_set {
	struct action	in;
	struct action	out;
	struct action	resolve;
};

struct rule {
	int	type;
	int	conform_policy;
	union {
		struct connection_rule	*connection;
		struct resolve_rule	*resolve;
		struct policy_set	policy;
	} rule;
};



static int
resolve_line_to_rule(struct rule *rule, char *line)
{
	char *node = NULL;
	char *service = NULL;
	char *spoof_node = NULL;
	char *spoof_service = NULL;
	struct action action;

	char *t0, *t1, *ptr;

	action_zero(&action);

	t0 = strtok_r(line, " ", &ptr);
	while (t0) {
		if (strncmp(t0, "node=", 5) == 0) {
			node = t0 + 5;
			t1 = strchr(node, '/');
			if (t1) {
				*t1 = '\0';
				service = t1 + 1;
			}
		} else if (strncmp(t0, "spoof=", 6) == 0) {
			spoof_node = t0 + 6;
			t1 = strchr(spoof_node, '/');
			if (t1) {
				*t1 = '\0';
				spoof_service = t1 + 1;
			}
		} else if (strncmp(t0, "ret=", 4) == 0) {
			action.ret = atoi(t0 + 4);
			action.set_ret = 1;
			rule->conform_policy = 0;
		} else if (strncmp(t0, "err=", 4) == 0) {
			action.err = atoi(t0 + 4);
			action.set_err = 1;
			rule->conform_policy = 0;
		} else {
			log_warn("unknown keyword: \"%s\"", t0);
		}

		t0 = strtok_r(NULL, " ", &ptr);
	}

	if (node || service || spoof_node || spoof_service) {
		rule->rule.resolve = rule_resolve_new(node, service,
				spoof_node, spoof_service,
				&action);
		rule->type = TYPE_RESOLVE;
		return 0;
	} else {
		return 1;
	}
}



static int
connection_line_to_rule(struct rule *rule, char *line) /* TODO spoof, protocol */
{
	char *from_addr = NULL;
	char *to_addr = NULL;
	uint16_t from_port = 0;
	uint16_t to_port = 0;
	struct action action;
	char *t0, *t1, *ptr;

	action_zero(&action);

	t0 = strtok_r(line, " ", &ptr);
	while (t0) {
		if (strncmp(t0, "src=", 4) == 0) {
			from_addr = t0 + 4;
			t1 = strchr(from_addr, '/');
			if (t1) {
				*t1 = '\0';
				from_port = atoi(t1 + 1);
			}
		} else if (strncmp(t0, "dest=", 5) == 0) {
			to_addr = t0 + 5;
			t1 = strchr(to_addr, '/');
			if (t1) {
				*t1 = '\0';
				to_port = atoi(t1 + 1);
			}
		} else if (strncmp(t0, "sport=", 6) == 0) {
			from_port = atoi(t0 + 6);
		} else if (strncmp(t0, "dport=", 6) == 0) {
			to_port = atoi(t0 + 6);
		} else if (strncmp(t0, "ret=", 4) == 0) {
			int err = atoi(t0 + 4);
			if (err) {
				action.ret = -1;
				action.err = err;
				action.set_ret = action.set_err = 1;
			}
			rule->conform_policy = 0;
		} else if (strcmp(t0, "drop") == 0) {
			action.ret = -1;
			action.err = ETIMEDOUT;
			action.set_ret = action.set_err = 1;
			rule->conform_policy = 0;
		} else if (strcmp(t0, "accept") == 0) {
			rule->conform_policy = 0;
		} else {
			log_warn("unknown keyword: \"%s\"", t0);
		}


		t0 = strtok_r(NULL, " ", &ptr);
	}

	if (from_addr || to_addr || from_port || to_port ||
			!rule->conform_policy) {
		rule->rule.connection = rule_connection_new(0,
				from_addr, from_port,
				to_addr, to_port,
				NULL, 0,
				&action);
		rule->type = TYPE_CONNECTION;
		return 0;
	} else {
		return 1;
	}
}



static void
set_policy(struct action *policy, char *str)
{
	char *sep;
	sep = strchr(str, '/');
	if (sep) {
		*sep = '\0';
		policy->err = atoi(sep + 1);
		policy->set_err = 1;
	}
	policy->ret = atoi(str);
	policy->set_ret = 1;
}



static int
policy_line_to_rule(struct rule *rule, char *line)
{
	char *param, *ptr;
	struct policy_set *policy = &(rule->rule.policy);
	int err = 0;

	memset(policy, 0, sizeof(rule->rule.policy));

	printf("--> %s <--\n", line);
	param = strtok_r(line, " ", &ptr);
	while (param) {
		if (strncmp(param, "in=", 3) == 0) {
			set_policy(&(policy->in), param + 3);
		} else if (strncmp(param, "out=", 4) == 0) {
			set_policy(&(policy->out), param + 4);
		} else if (strncmp(param, "resolve=", 8) == 0) {
			set_policy(&(policy->resolve), param + 8);
		} else {
			log_warn("unknown keyword: \"%s\"", param);
			err = 1;
		}

		param = strtok_r(NULL, " ", &ptr);
	}

	return !err;
}



static void
line_to_rule(struct rule *rule, char *line)
{
	while (*line == ' ') {
		line++;
	}
	rule->type = TYPE_INVALID;

	if (strncmp(line, "resolve ", 8) == 0) {
		resolve_line_to_rule(rule, line + 8);
	} else if (strncmp(line, "in ", 3) == 0) {
		if (connection_line_to_rule(rule, line + 3) == 0) {
			rule->type = TYPE_CONNECTION_INBOUND;
		}
	} else if (strncmp(line, "out ", 4) == 0) {
		if (connection_line_to_rule(rule, line + 4) == 0) {
			rule->type = TYPE_CONNECTION_OUTBOUND;
		}
	} else if (strncmp(line, "policy ", 7) == 0) {
		if (policy_line_to_rule(rule, line + 7)) {
			rule->type = TYPE_POLICY;
		}
	}
}



static void
add_line_to_rules(struct rules *rules, char *line)
{
	struct rule rule;
	rule.conform_policy = 1;
	line_to_rule(&rule, line);
	switch (rule.type) {
		case TYPE_CONNECTION_INBOUND:
			if (rule.conform_policy) {
				action_copy(&(rule.rule.connection->policy),
						&(rules->in_policy));
			}
			rules_add_inbound(rules, rule.rule.connection);
			break;
		case TYPE_CONNECTION_OUTBOUND:
			if (rule.conform_policy) {
				action_copy(&(rule.rule.connection->policy),
						&(rules->out_policy));
			}
			rules_add_outbound(rules, rule.rule.connection);
			break;
		case TYPE_RESOLVE:
			if (rule.conform_policy) {
				action_copy(&(rule.rule.resolve->policy),
						&(rules->resolve_policy));
			}
			rules_add_resolve(rules, rule.rule.resolve);
			break;
		case TYPE_POLICY:
			if (rule.rule.policy.in.set_ret ||
					rule.rule.policy.in.set_err) {
				action_copy(&(rules->in_policy),
						&(rule.rule.policy.in));
			}
			if (rule.rule.policy.out.set_ret ||
					rule.rule.policy.out.set_err) {
				action_copy(&(rules->out_policy),
						&(rule.rule.policy.out));
			}
			if (rule.rule.policy.resolve.set_ret ||
					rule.rule.policy.resolve.set_err) {
				action_copy(&(rules->out_policy),
						&(rule.rule.policy.resolve));
			}
			break;

		case TYPE_INVALID:
		default:
			log_warn("invalid rule: \"%s\"", line);
			break;
	};
}



struct rules *
parser_buffer_to_rules(const char *str)
{
	struct rules *rules;
	char line[RULE_LEN + 1];
	const char *start, *end, *end_a, *end_b;
	size_t len;

	rules = rules_new();
	line[RULE_LEN] = '\0';

	start = str;
	while (*start) {
		end_a = strchr(start, '\n');
		end_b = strchr(start, ';');
		if (end_a == NULL && end_b == NULL) {
			strncpy(line, start, RULE_LEN);
			add_line_to_rules(rules, line);
			break;
		} else if (end_a && end_b == NULL) {
			end = end_a;
		} else if (end_b && end_a == NULL) {
			end = end_b;
		} else if (end_a < end_b) {
			end = end_a;
		} else {
			end = end_b;
		}
		len = end - start;
		if (len > RULE_LEN) {
			len = RULE_LEN;
		}
		strncpy(line, start, len);
		line[len] = '\0';
		add_line_to_rules(rules, line);
		start = end + 1;
	}

	return rules;
}



#ifdef APP
int
main(int argc, char **argv)
{
	struct rules *rules = NULL;
	rules = parser_buffer_to_rules(argv[1]);
	rules_to_log(rules);
	rules_free(rules);
	return 0;
}
#endif
